/*
  Copyright (C) 2017 Paul Davis

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef __ardour_selection_h__
#define __ardour_selection_h__

#include <set>
#include <vector>

#include <boost/weak_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "pbd/stateful.h"
#include "pbd/i18n.h"

#include "ardour/presentation_info.h"

namespace ARDOUR {

class AutomationControl;
class Session;
class Stripable;
class PresentationInfo;

class LIBARDOUR_API CoreSelection : public PBD::Stateful {
  public:
	CoreSelection (Session& s);
	~CoreSelection ();

	void toggle (boost::shared_ptr<Stripable>, boost::shared_ptr<AutomationControl>);
	void add (boost::shared_ptr<Stripable>, boost::shared_ptr<AutomationControl>);
	void remove (boost::shared_ptr<Stripable>, boost::shared_ptr<AutomationControl>);
	void set (boost::shared_ptr<Stripable>, boost::shared_ptr<AutomationControl>);
	void clear_stripables();

	bool selected (boost::shared_ptr<const Stripable>) const;
	bool selected (boost::shared_ptr<const AutomationControl>) const;
	uint32_t selected() const;

	struct StripableAutomationControl {
		boost::shared_ptr<Stripable> stripable;
		boost::shared_ptr<AutomationControl> controllable;
		int order;

		StripableAutomationControl (boost::shared_ptr<Stripable> s, boost::shared_ptr<AutomationControl> c, int o)
			: stripable (s), controllable (c), order (o) {}
	};

	typedef std::vector<StripableAutomationControl> StripableAutomationControls;

	void get_stripables (StripableAutomationControls&) const;

	XMLNode& get_state (void);
	int set_state (const XMLNode&, int version);

  protected:
	friend class AutomationControl;
	void remove_control_by_id (PBD::ID const &);

  protected:
	friend class Stripable;
	friend class Session;
	void remove_stripable_by_id (PBD::ID const &);

  private:
	mutable Glib::Threads::RWLock _lock;
	Session& session;
	int selection_order;

	struct SelectedStripable {
		SelectedStripable (boost::shared_ptr<Stripable>, boost::shared_ptr<AutomationControl>, int);
		SelectedStripable (PBD::ID const & s, PBD::ID const & c, int o)
			: stripable (s), controllable (c), order (o) {}

		PBD::ID stripable;
		PBD::ID controllable;
		int order;

		bool operator< (SelectedStripable const & other) const {
			if (stripable == other.stripable) {
				return controllable < other.controllable;
			}
			return stripable < other.stripable;
		}
	};

	typedef std::set<SelectedStripable> SelectedStripables;

	SelectedStripables _stripables;

	void send_selection_change ();
};

} // namespace ARDOUR

#endif /* __ardour_selection_h__ */
